function [E,w,b] = cnnfit(x,y,w,b,n)
% Author: P. Bhattacharya
%
%--------------------------------------------------------------------------
% This source code is provided without any warranties as published in 
% DASP book 3rd edition, copyright Wiley & Sons 2022, available at 
% https://mydasp.com/dasp. It may be used for educational purposes and not 
% for commercial applications without further permission.
%--------------------------------------------------------------------------
% dasp_chap11_ex2_CNN -> cnnfit.m
% May 04, 2022, Purbaditya Bhattacharya
% Exercise for Chapter "Machine Learning for Audio"

L = length(x);
numlayers = length(w);
%% placeholder for hidden layers
h1 = cell(1,numlayers);
h2 = cell(1,numlayers-1);
dh2 = cell(1,numlayers-1);
dh2h1 = cell(1,numlayers-1);
%% train and update
for i = 1:L-n+1
    %% get data snippet
    x_i = x(i:i+n-1);
    y_i = y(i+n-1);
    %% feedforward operation
    for j = 1:numlayers
        if j == 1
            % convolution with zero padding
            h1{j} = convl(x_i,w{j},b{j});
            % h1{j} vectorized before ReLU
            h2{j} = max(h1{j}(:)',0);
        elseif j == numlayers
            h1{j} = h2{j-1}*w{j}+b{j};
        else
            h1{j} = h2{j-1}*w{j}+b{j};
            h2{j} = max(h1{j},0);
        end
    end
    %% error and backpropagation
    de = -2*(y_i-h1{end});
    E = sum(de.^2);
    for j = numlayers:-1:1
        if j == numlayers
            dh2{j-1} = de*w{j}';
            dw{j} = h2{j-1}'*de;
            db{j} = de;
        elseif j == 1
            dh2h1{j} = h2{j};dh2h1{j}(dh2h1{j}>0)=1;
            dh1 = reshape((dh2{j}.*dh2h1{j}),size(x_i,1),size(x_i,2),size(w{j},4));
            [~,dw{j},db{j}] = dconvl(x_i,w{j},b{j},dh1);
        else
            dh2h1{j} = h2{j};dh2h1{j}(dh2h1{j}>0)=1;
            dh2{j-1} = (dh2{j}.*dh2h1{j})*w{j}';
            dw{j} = h2{j-1}'*(dh2{j}.*dh2h1{j});
            db{j} = dh2{j}.*dh2h1{j};
        end
        
        %% update SGD
        w{j} = w{j}-0.01*dw{j};
        b{j} = b{j}-0.001*db{j};
    end
end
end

%% filter function
function y = convl(x,w,b)
    y = zeros(size(x,1),size(x,2),size(w,4));
    for d = 1:size(w,4)
        for ch = 1:size(x,3)
            y(:,:,d) = y(:,:,d) + conv(x(:,:,ch),rot90(w(:,:,ch,d),2),'same') + b(d);
        end
    end
end

%% derivative of the filter function
function [dx,dw,db] = dconvl(x,w,b,dy)
dw = zeros(size(w));
db = zeros(size(b));
dx = zeros(size(x));
x_p = padding(x,w);
for d = 1:size(w,4)
    for ch = 1:size(w,3)
        dw(:,:,ch,d) = conv(x_p(:,:,ch),rot90(dy(:,:,d),2),'valid');
        dx(:,:,ch) = dx(:,:,ch)+conv(dy(:,:,d),w(:,:,ch,d),'same');
    end
    db(d) = sum(dy(:,:,d),'all');
end
end

%% Padding function
function y = padding(x,w)
    pady = floor(size(w,1)/2);
    padx = floor(size(w,2)/2);
    y = padarray(x,[padx,pady,0],0);
end