% Author: P. Bhattacharya
%
%--------------------------------------------------------------------------
% This source code is provided without any warranties as published in 
% DASP book 3rd edition, copyright Wiley & Sons 2022, available at 
% https://mydasp.com/dasp. It may be used for educational purposes and not 
% for commercial applications without further permission.
%--------------------------------------------------------------------------
% dasp_chap11_ex1_NN
% May 04, 2022, Purbaditya Bhattacharya
% Exercise for Chapter "Machine Learning for Audio"

clear all;close all;clc

%% generate datapoints to fit
numdatapoints = 1000;
x = 2*(rand(1,numdatapoints)-0.5);
y = max(sin(x),-0.2);
figure(1);scatter([1:length(y)],y);grid on;xlabel('Samples');ylabel('Output');

%% number of hidden layers and input/hidden layer dimension
numhiddenlayers = 1;           % number of hidden layers (excludes input and output layers)
n = 1;                         % input layer size
k = 5*ones(1,numhiddenlayers); % hidden layer size

%% number of epochs
N = 50;

%% definition and initialization of training parameters
w = cell(1,numhiddenlayers+1); % 
b = cell(1,numhiddenlayers+1);
for i = 1:numhiddenlayers+1
    if i == 1
        %% input->hidden
        w{i} = (2*rand(n,k(i))-1)/(n*k(i));
        b{i} = zeros(1,k(i));
    elseif i == numhiddenlayers+1
        %% hidden->output
        w{i} = (2*rand(k(i-1),1)-1)/(k(i-1)*1);
        b{i} = zeros(1,1);
    else
        %% hidden->hidden
        w{i} = (2*rand(k(i-1),k(i))-1)/(k(i-1)*k(i));
        b{i} = zeros(1,k(i));
    end
end
E = zeros(1,N);

%% fit a model
for i = 1:N
    [E(i),w,b] = functions.nnfit(x,y,w,b,n);
end

%% plot the error curve
figure(2);plot(E);grid on;xlabel('Epoch');ylabel('Error');

%% test the model
x = -1:0.001:1;
y = max(sin(x),-0.2);
y_est = functions.nntest(x,w,b,n);
figure(3);plot(x(1:length(y_est)),y(1:length(y_est)));hold on;plot(x(1:length(y_est)),y_est,'r--');grid on;hold off;legend('ground truth','estimated');xlabel('Input');ylabel('Output');

disp(mean((y(1:length(y_est))-y_est).^2));