% Author: P. Bhattacharya
%
%--------------------------------------------------------------------------
% This source code is provided without any warranties as published in 
% DASP book 3rd edition, copyright Wiley & Sons 2022, available at 
% https://mydasp.com/dasp. It may be used for educational purposes and not 
% for commercial applications without further permission.
%--------------------------------------------------------------------------
% dasp_chap11_ex2_CNN
% May 04, 2022, Purbaditya Bhattacharya
% Exercise for Chapter "Machine Learning for Audio"

clear all;close all;clc

%% generate datapoints to fit
numdatapoints = 4000;
x = 2*(rand(1,numdatapoints)-0.5);
y = max(sin(x),-0.2);
figure(1);scatter([1:length(y)],y);grid on;xlabel('Samples');ylabel('Output');

%% number of fully connected (fc) layers and input/hidden layer dimensions
numfclayers = 1;             % number of hidden layers (fc), does not include input (conv) and output layer.
n = 5;                       % input layer size
k = 2*ones(1,numfclayers);   % hidden layer size
d = 3;                       % num filter groups for conv (Filter group size - 3x1x1x3)

%% number of epochs
N = 250;

%% definition and initialization of training parameters
w = cell(1,numfclayers+2); % 
b = cell(1,numfclayers+2);
for i = 1:numfclayers+2
    if i == 1
        %% input -> conv (hidden)
        w{i} = (2*rand(min(n,3),1,1,d)-1)/(min(n,3)*d);
        b{i} = zeros(1,3);
    elseif i == numfclayers+2
        %% fc (hidden) -> output
        w{i} = (2*rand(k(i-2),1)-1)/(k(i-2)*1);
        b{i} = zeros(1,1);
    elseif i == 2
        %% conv (hidden) -> fc (hidden)
        w{i} = (2*rand(n*d,k(i-1))-1)/(n*d*k(i-1));
        b{i} = zeros(1,k(i-1));
    else
        %% fc (hidden) -> fc (hidden) (if numfclayers > 1)
        w{i} = (2*rand(k(i-2),k(i-1))-1)/(k(i-2)*k(i-1));
        b{i} = zeros(1,k(i-1));
    end
end
E = zeros(1,N);

%% fit a model
for i = 1:N
    [E(i),w,b] = functions.cnnfit(x,y,w,b,n);
end

%% plot the error curve
figure(2);plot(E);grid on;xlabel('Epoch');ylabel('Error');

%% test the model
x = -1:0.001:1;
y = max(sin(x),-0.2);
y_est = functions.cnntest(x,w,b,n);
figure(3);plot(x(1:length(y_est)),y(1:length(y_est)));hold on;plot(x(1:length(y_est)),y_est,'r--');grid on;hold off;legend('ground truth','estimated');xlabel('Input');ylabel('Output');

disp(mean((y(1:length(y_est))-y_est).^2));